@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2019-2024 by Sukchan Lee <acetcom@gmail.com>
+ * Copyright (C) 2019-2025 by Sukchan Lee <acetcom@gmail.com>
  *
  * This file is part of Open5GS.
  *
@@ -30,6 +30,8 @@
 #include "namf-handler.h"
 #include "nsmf-handler.h"
 #include "npcf-handler.h"
+#include "nsmf-handler.h"
+#include "binding.h"
 
 void smf_state_initial(ogs_fsm_t *s, smf_event_t *e)
 {
@@ -73,6 +75,8 @@ void smf_state_operational(ogs_fsm_t *s, smf_event_t *e)
     ogs_sbi_stream_t *stream = NULL;
     ogs_pool_id_t stream_id = OGS_INVALID_POOL_ID;
     ogs_sbi_request_t *sbi_request = NULL;
+    int state = 0;
+    bool unknown_res_status = false;
 
     ogs_sbi_nf_instance_t *nf_instance = NULL;
     ogs_sbi_subscription_data_t *subscription_data = NULL;
@@ -550,7 +554,7 @@ void smf_state_operational(ogs_fsm_t *s, smf_event_t *e)
                         break;
 
                     DEFAULT
-                        sess = smf_sess_add_by_sbi_message(&sbi_message);
+                        sess = smf_sess_add_by_sm_context(&sbi_message);
                         if (!sess) {
                             ogs_error("smf_sess_add_by_sbi_message() failed");
                             smf_sbi_send_sm_context_create_error(stream,
@@ -587,6 +591,124 @@ void smf_state_operational(ogs_fsm_t *s, smf_event_t *e)
                 }
                 break;
 
+            CASE(OGS_SBI_RESOURCE_NAME_PDU_SESSIONS)
+                SWITCH(sbi_message.h.method)
+                CASE(OGS_SBI_HTTP_METHOD_POST)
+                    SWITCH(sbi_message.h.resource.component[2])
+                    CASE(OGS_SBI_RESOURCE_NAME_MODIFY)
+                    CASE(OGS_SBI_RESOURCE_NAME_RELEASE)
+                        if (!sbi_message.h.resource.component[1]) {
+                            ogs_error("No pduSessionRef [%s]",
+                                    sbi_message.h.resource.component[1]);
+                            smf_sbi_send_hsmf_update_error(stream,
+                                    OGS_SBI_HTTP_STATUS_BAD_REQUEST,
+                                    OGS_SBI_APP_ERRNO_NULL,
+                                    OGS_5GSM_CAUSE_INVALID_MANDATORY_INFORMATION,
+                                    "No pduSessionRef",
+                                    sbi_message.h.resource.component[1], NULL);
+                            break;
+                        }
+
+                        sess = smf_sess_find_by_pdu_session_ref(
+                                sbi_message.h.resource.component[1]);
+
+                        if (!sess) {
+                            ogs_warn("Not found [%s]", sbi_message.h.uri);
+                            smf_sbi_send_hsmf_update_error(stream,
+                                    OGS_SBI_HTTP_STATUS_BAD_REQUEST,
+                                    OGS_SBI_APP_ERRNO_NULL,
+                                    OGS_5GSM_CAUSE_INVALID_MANDATORY_INFORMATION,
+                                    "Not found", sbi_message.h.uri, NULL);
+                        }
+                        break;
+
+                    DEFAULT
+                        sess = smf_sess_add_by_pdu_session(&sbi_message);
+                        if (!sess) {
+                            ogs_error("smf_sess_add_by_sbi_message() failed");
+                            smf_sbi_send_pdu_session_create_error(stream,
+                                OGS_SBI_HTTP_STATUS_BAD_REQUEST,
+                                OGS_SBI_APP_ERRNO_NULL,
+                                OGS_5GSM_CAUSE_INVALID_MANDATORY_INFORMATION,
+                                "smf_sess_add_by_sbi_message() failed",
+                                NULL, NULL);
+                            break;
+                        }
+
+                        smf_metrics_inst_by_slice_add(NULL, NULL,
+                                SMF_METR_CTR_SM_PDUSESSIONCREATIONREQ, 1);
+                    END
+                    break;
+
+                DEFAULT
+                    ogs_error("Invalid HTTP method [%s]", sbi_message.h.method);
+                    ogs_assert(true ==
+                        ogs_sbi_server_send_error(stream,
+                            OGS_SBI_HTTP_STATUS_BAD_REQUEST, &sbi_message,
+                            "Invalid HTTP method", sbi_message.h.method,
+                            NULL));
+                END
+
+                if (sess) {
+                    smf_ue = smf_ue_find_by_id(sess->smf_ue_id);
+                    ogs_assert(smf_ue);
+                    ogs_assert(OGS_FSM_STATE(&sess->sm));
+
+                    e->sess_id = sess->id;
+                    e->h.sbi.message = &sbi_message;
+                    ogs_fsm_dispatch(&sess->sm, e);
+                }
+                break;
+
+            CASE(OGS_SBI_RESOURCE_NAME_VSMF_PDU_SESSIONS)
+                SWITCH(sbi_message.h.method)
+                CASE(OGS_SBI_HTTP_METHOD_POST)
+                    if (!sbi_message.h.resource.component[1]) {
+                        ogs_error("No smContextRef [%s]",
+                                sbi_message.h.resource.component[1]);
+                        smf_sbi_send_vsmf_update_error(stream,
+                                OGS_SBI_HTTP_STATUS_BAD_REQUEST,
+                                OGS_SBI_APP_ERRNO_NULL,
+                                OGS_5GSM_CAUSE_INVALID_MANDATORY_INFORMATION,
+                                "No smContextRef",
+                                sbi_message.h.resource.component[1], NULL);
+                        break;
+                    }
+
+                    sess = smf_sess_find_by_sm_context_ref(
+                            sbi_message.h.resource.component[1]);
+
+                    if (!sess) {
+                        ogs_warn("Not found [%s]", sbi_message.h.uri);
+                        smf_sbi_send_vsmf_update_error(stream,
+                                OGS_SBI_HTTP_STATUS_BAD_REQUEST,
+                                OGS_SBI_APP_ERRNO_NULL,
+                                OGS_5GSM_CAUSE_INVALID_MANDATORY_INFORMATION,
+                                "Not found", sbi_message.h.uri, NULL);
+                    }
+                    break;
+
+                DEFAULT
+                    ogs_error("Invalid HTTP method [%s]", sbi_message.h.method);
+                    ogs_assert(true ==
+                        ogs_sbi_server_send_error(stream,
+                            OGS_SBI_HTTP_STATUS_BAD_REQUEST, &sbi_message,
+                            "Invalid HTTP method", sbi_message.h.method,
+                            NULL));
+                    break;
+                END
+
+                if (sess) {
+                    smf_ue = smf_ue_find_by_id(sess->smf_ue_id);
+                    ogs_assert(smf_ue);
+                    ogs_assert(OGS_FSM_STATE(&sess->sm));
+
+                    e->sess_id = sess->id;
+                    e->h.sbi.message = &sbi_message;
+                    ogs_fsm_dispatch(&sess->sm, e);
+                }
+                break;
+
             DEFAULT
                 ogs_error("Invalid resource name [%s]",
                         sbi_message.h.resource.component[0]);
@@ -631,10 +753,13 @@ void smf_state_operational(ogs_fsm_t *s, smf_event_t *e)
 
                 SWITCH(sbi_message.h.resource.component[2])
                 CASE(OGS_SBI_RESOURCE_NAME_UPDATE)
+#define ENABLE_PCF_INITIATED_SESSION_RELEASE 0
+#if !ENABLE_PCF_INITIATED_SESSION_RELEASE
                     smf_npcf_smpolicycontrol_handle_update_notify(
                             sess, stream, &sbi_message);
                     break;
                 CASE(OGS_SBI_RESOURCE_NAME_TERMINATE)
+#endif
                     smf_npcf_smpolicycontrol_handle_terminate_notify(
                             sess, stream, &sbi_message);
                     break;
@@ -836,6 +961,7 @@ void smf_state_operational(ogs_fsm_t *s, smf_event_t *e)
         CASE(OGS_SBI_SERVICE_NAME_NUDM_SDM)
         CASE(OGS_SBI_SERVICE_NAME_NPCF_SMPOLICYCONTROL)
         CASE(OGS_SBI_SERVICE_NAME_NAMF_COMM)
+        CASE(OGS_SBI_SERVICE_NAME_NSMF_PDUSESSION)
             sbi_xact_id = OGS_POINTER_TO_UINT(e->h.sbi.data);
             ogs_assert(sbi_xact_id >= OGS_MIN_POOL_ID &&
                     sbi_xact_id <= OGS_MAX_POOL_ID);
@@ -877,9 +1003,6 @@ void smf_state_operational(ogs_fsm_t *s, smf_event_t *e)
             break;
 
         CASE(OGS_SBI_SERVICE_NAME_NUDM_UECM)
-            int state = 0;
-            bool unknown_res_status = false;
-
             sbi_xact_id = OGS_POINTER_TO_UINT(e->h.sbi.data);
             ogs_assert(sbi_xact_id >= OGS_MIN_POOL_ID &&
                     sbi_xact_id <= OGS_MAX_POOL_ID);
@@ -914,13 +1037,16 @@ void smf_state_operational(ogs_fsm_t *s, smf_event_t *e)
             smf_ue = smf_ue_find_by_id(sess->smf_ue_id);
             ogs_assert(smf_ue);
 
-            if (state == SMF_UECM_STATE_REGISTERED) {
+            if (state == SMF_UECM_STATE_REGISTERED ||
+                state == SMF_UECM_STATE_REGISTERED_HR) {
                 /* SMF Registration */
                 if (sbi_message.res_status != OGS_SBI_HTTP_STATUS_OK &&
                     sbi_message.res_status != OGS_SBI_HTTP_STATUS_CREATED)
                     unknown_res_status = true;
-            } else if (state == SMF_UECM_STATE_DEREGISTERED_BY_AMF ||
-                        state == SMF_UECM_STATE_DEREGISTERED_BY_N1_N2_RELEASE) {
+            } else if (state == SMF_UECM_STATE_DEREG_BY_AMF ||
+                        state == SMF_UECM_STATE_DEREG_BY_AMF_HR ||
+                        state == SMF_UECM_STATE_DEREG_BY_N1N2 ||
+                        state == SMF_UECM_STATE_DEREG_BY_N1N2_HR) {
                 /* SMF Deregistration */
                 if (sbi_message.res_status != OGS_SBI_HTTP_STATUS_NO_CONTENT)
                     unknown_res_status = true;
@@ -946,18 +1072,169 @@ void smf_state_operational(ogs_fsm_t *s, smf_event_t *e)
             }
 
             if (state == SMF_UECM_STATE_REGISTERED) {
-                /* SMF Registration */
                 ogs_assert(stream);
                 ogs_assert(true == ogs_sbi_send_http_status_no_content(stream));
-            } else if (state == SMF_UECM_STATE_DEREGISTERED_BY_AMF) {
+                smf_metrics_inst_by_slice_add(
+                        &sess->serving_plmn_id, &sess->s_nssai,
+                        SMF_METR_CTR_SM_PDUSESSIONCREATIONSUCC, 1);
+            } else if (state == SMF_UECM_STATE_REGISTERED_HR) {
+                ogs_assert(stream);
+                smf_sbi_send_pdu_session_created_data(sess, stream);
+                smf_metrics_inst_by_slice_add(
+                        &sess->serving_plmn_id, &sess->s_nssai,
+                        SMF_METR_CTR_SM_PDUSESSIONCREATIONSUCC, 1);
+
+/*
+ * Non-roaming/LBO: start network-triggered PDU Session Modification at step 11
+ * after N1N2 transfer (Establishment Accept) and N2/N4 context sync, ensuring
+ * the session is active on UE, RAN, and SMF before applying QoS updates.
+ *
+ * Home-Routed Roaming: trigger PDU Session Modification at step 13
+ * immediately after H-SMF’s CreateSMContext response and H-UPF N4 setup
+ * to apply QoS updates without waiting for V-SMF or RAN setup.
+ */
+                smf_qos_flow_binding(sess);
+
+            } else if (state == SMF_UECM_STATE_DEREG_BY_AMF) {
+/*
+ * Handle deregistration states SMF_UECM_STATE_DEREG_BY_AMF and
+ * SMF_UECM_STATE_DEREG_BY_N1N2: context has been removed by AMF or
+ * by NAS (N1/N2) procedures, respectively.
+ *
+ * In 3GPP TS 23.501, two related flows must be coordinated:
+ *  - PDU Session Release: Figure 4.3.4.2-1, steps 11 to 15
+ *  - PDU Session Establishment: Figure 4.3.2.2.1-1, steps 16a to 16c
+ *
+ * Normally, step 11 (SMContextStatusNotify) informs the AMF that the SM
+ * Context has been released, and SMF would then delete the session state.
+ * However, if the AMF triggers a release while a new establishment is still
+ * underway (the PFCP Modification and NAS Registration in steps 16a–16c),
+ * deleting the context too early causes those procedures to fail.
+ *
+ * To avoid this race, we defer step 11 until after steps 12–15 complete:
+ * subscription termination, policy cleanup, and UDM deregistration. This
+ * delay allows both the old (released) context and the new (establishing)
+ * context to coexist in SMF, so that PFCP and NAS messages can still find
+ * the correct context. Only once all cleanup is done do we invoke the
+ * SMContextStatusNotify and remove the session state.
+ *
+ * This ensures:
+ *  - PFCP Modification (step 16a) still locates its context
+ *  - NAS Registration (steps 16b–16c) can finish successfully
+ *  - The final cleanup (deferred step 11) does not interrupt any in-flight
+ *    procedures
+ *
+ * If you attempt to address this:
+ *
+ * Legacy flow:
+ *   sess = smf_sess_find_by_psi(...);
+ *   if (sess) { smf_sess_remove(sess); }
+ *   sess = smf_sess_add_by_psi(...);
+ *
+ *  - Do not use smf_sess_find_by_psi() to remove old contexts here.
+ *  - Call only smf_sess_add_by_psi() to add the new session.
+ *  - Since PSI duplicates may occur without an explicit release, you must
+ *    design a separate cleanup mechanism (e.g., periodic purge, release
+ *    tracking, or context deduplication) that safely detects and removes
+ *    stale SM Context entries outside of this path.
+ */
                 /* SMF Deregistration */
-                ogs_assert(stream);
-                ogs_assert(true == ogs_sbi_send_http_status_no_content(stream));
+                if (stream)
+                    ogs_assert(true ==
+                            ogs_sbi_send_http_status_no_content(stream));
+                else
+                    ogs_error("Stream has already been removed");
+
                 SMF_SESS_CLEAR(sess);
-            } else if (state == SMF_UECM_STATE_DEREGISTERED_BY_N1_N2_RELEASE) {
-                /* SMF Deregistration */
+            } else if (state == SMF_UECM_STATE_DEREG_BY_N1N2) {
                 ogs_assert(true == smf_sbi_send_sm_context_status_notify(sess));
                 SMF_SESS_CLEAR(sess);
+            } else if (state == SMF_UECM_STATE_DEREG_BY_AMF_HR) {
+    /*
+     * Network-requested PDU Session Release(DUPLICATED)
+     *
+     * 1.  V: OGS_PFCP_MODIFY_HOME_ROUTED_ROAMING|OGS_PFCP_MODIFY_UL_ONLY|
+     *        OGS_PFCP_MODIFY_DEACTIVATE
+     * 2.  V: OGS_PFCP_DELETE_TRIGGER_AMF_UPDATE_SM_CONTEXT,
+     * 3.  V: OpenAPI_request_indication_NW_REQ_PDU_SES_REL
+     * 4.  V: smf_nsmf_pdusession_build_hsmf_update_data
+     * 5.  H: smf_nsmf_handle_update_data_in_hsmf
+     * 6.  H: OpenAPI_request_indication_NW_REQ_PDU_SES_REL
+     * 6.  H: e->h.sbi.state = OGS_PFCP_DELETE_TRIGGER_AMF_UPDATE_SM_CONTEXT
+     * 7.  H: ogs_sbi_send_http_status_no_content
+     * 8.  H: OGS_FSM_TRAN(s, smf_gsm_state_wait_pfcp_deletion)
+     * 9.  H: smf_sbi_cleanup_session(SMF_UECM_STATE_DEREG_BY_AMF_HR
+     *                                SMF_SBI_CLEANUP_MODE_POLICY_FIRST);
+     * 10. H: OGS_FSM_TRAN(s, smf_gsm_state_5gc_session_will_deregister);
+     * 11. H*: SMF_SESS_CLEAR(sess)
+     * 12. V: e->h.sbi.state = OGS_PFCP_DELETE_TRIGGER_AMF_UPDATE_SM_CONTEXT
+     * 13. V: OGS_FSM_TRAN(s, smf_gsm_state_wait_pfcp_deletion)
+     * 14. V: ogs_sbi_send_http_status_no_content
+     * 15. V: OGS_FSM_TRAN(s, smf_gsm_state_session_will_release);
+     *
+     * Network-requested PDU Session Release
+     *
+     * 1.  V: smf_nsmf_handle_release_sm_context
+     * 2.  V: OGS_PFCP_MODIFY_HOME_ROUTED_ROAMING|OGS_PFCP_MODIFY_UL_ONLY|
+     *        OGS_PFCP_MODIFY_DEACTIVATE
+     * 3.  V: OGS_PFCP_DELETE_TRIGGER_AMF_RELEASE_SM_CONTEXT
+     * 4.  V: smf_nsmf_pdusession_build_release_data
+     * 5.  H: smf_nsmf_handle_release_data_in_hsmf
+     * 6.  H: e->h.sbi.state = OGS_PFCP_DELETE_TRIGGER_AMF_RELEASE_SM_CONTEXT
+     * 7.  H: OGS_FSM_TRAN(s, smf_gsm_state_wait_pfcp_deletion)
+     * 8.  H: ogs_sbi_send_http_status_no_content
+     * 9.  H: smf_sbi_cleanup_session(SMF_UECM_STATE_DEREG_BY_AMF_HR
+     *                                SMF_SBI_CLEANUP_MODE_POLICY_FIRST);
+     * 10. H: OGS_FSM_TRAN(s, smf_gsm_state_5gc_session_will_deregister);
+     * 11. H*: SMF_SESS_CLEAR(sess)
+     * 12. V: smf_nsmf_handle_release_data_in_hsmf
+     * 13. V: e->h.sbi.state = OGS_PFCP_DELETE_TRIGGER_AMF_RELEASE_SM_CONTEXT
+     * 14. V: OGS_FSM_TRAN(s, smf_gsm_state_wait_pfcp_deletion)
+     * 15. V: ogs_sbi_send_http_status_no_content
+     * 16. V: OGS_FSM_TRAN(s, smf_gsm_state_session_will_release);
+     */
+                SMF_SESS_CLEAR(sess);
+            } else if (state == SMF_UECM_STATE_DEREG_BY_N1N2_HR) {
+    /*
+     * UE-requested PDU Session Release
+     *
+     * 1.  V: OGS_PFCP_MODIFY_HOME_ROUTED_ROAMING|OGS_PFCP_MODIFY_UL_ONLY|
+     *        OGS_PFCP_MODIFY_DEACTIVATE
+     * 2.  V: OGS_PFCP_DELETE_TRIGGER_UE_REQUESTED
+     * 3.  V: OpenAPI_request_indication_UE_REQ_PDU_SES_REL
+     * 4.  V: smf_nsmf_pdusession_build_hsmf_update_data
+     * 5.  H: smf_nsmf_handle_update_data_in_hsmf
+     * 6.  H: OpenAPI_request_indication_UE_REQ_PDU_SES_REL
+     * 6.  H: e->h.sbi.state = OGS_PFCP_DELETE_TRIGGER_UE_REQUESTED
+     * 7.  H: ogs_sbi_send_http_status_no_content
+     * 8.  H: OGS_FSM_TRAN(s, smf_gsm_state_wait_pfcp_deletion)
+     * 9.  H: smf_nsmf_pdusession_build_vsmf_update_data
+     * 10. H: OGS_FSM_TRAN(s, smf_gsm_state_wait_5gc_n1_n2_release);
+     * 11. V: smf_nsmf_handle_update_data_in_vsmf
+     * 12. V: OpenAPI_request_indication_UE_REQ_PDU_SES_REL
+     * 13. V: e->h.sbi.state = OGS_PFCP_DELETE_TRIGGER_UE_REQUESTED
+     * 14. V: OGS_FSM_TRAN(s, smf_gsm_state_wait_pfcp_deletion)
+     * 15. V: ngap_build_pdu_session_resource_release_command_transfer+
+     *        gsm_build_pdu_session_release_command
+     * 16  V: OGS_FSM_TRAN(&sess->sm, smf_gsm_state_wait_5gc_n1_n2_release)
+     * 17. V: ogs_sbi_send_http_status_no_content(stream)
+     * 18. V: case OpenAPI_n2_sm_info_type_PDU_RES_REL_RSP:
+     *        case OGS_NAS_5GS_PDU_SESSION_RELEASE_COMPLETE:
+     * 19. V: ogs_sbi_send_http_status_no_content(n1_n2_released_stream)
+     * 20. V: OGS_FSM_TRAN(s, smf_gsm_state_5gc_session_will_deregister);
+     * 21. H: case OGS_EVENT_SBI_CLIENT:
+     * 22. H: CASE(OGS_SBI_RESOURCE_NAME_VSMF_PDU_SESSIONS)
+     * 23. H: smf_sbi_cleanup_session(SMF_UECM_STATE_DEREG_BY_N1N2_HR
+     *                                SMF_SBI_CLEANUP_MODE_POLICY_FIRST);
+     * 24. H*: smf_sbi_send_status_notify+SMF_SESS_CLEAR(sess)
+     * 25. V: case OGS_EVENT_SBI_SERVER:
+     * 26. V: CASE(OGS_SBI_RESOURCE_NAME_VSMF_PDU_SESSIONS)
+     * 27. V: ogs_sbi_send_http_status_no_content+
+     *        smf_sbi_send_sm_context_status_notify
+     * 28. V: OGS_FSM_TRAN(s, smf_gsm_state_session_will_release);
+     */
+                ogs_assert(true == smf_sbi_send_status_notify(sess));
+                SMF_SESS_CLEAR(sess);
             }
 
             break;
@@ -1137,6 +1414,16 @@ void smf_state_operational(ogs_fsm_t *s, smf_event_t *e)
         ogs_pkbuf_free(pkbuf);
         break;
 
+    case SMF_EVT_SESSION_RELEASE:
+        sess = smf_sess_find_by_id(e->sess_id);
+        if (!sess) {
+            ogs_error("Session has already been removed");
+            break;
+        }
+
+        ogs_fsm_dispatch(&sess->sm, e);
+        break;
+
     default:
         ogs_error("No handler for event %s", smf_event_get_name(e));
         break;
